import sys
from enum import Enum
from math import ceil
from typing import List
from xml.dom.minidom import Element

from sqlalchemy import text

from app.lib import util

""" A 'component' to render a parametised query to a paginated and tabulated view """

class FieldType(Enum):
    """ to identify how a field is intended to be used """
    STRING = 10
    DATE = 20
    LOOKUP = 30


class Field:
    """ encapsulates the information needed to describe a field (column) in a grid
        parameters:
            field(str): the name of the field in the db
            label(str): A human readble label for the field (column header)
            mode(FieldType) the datatype for the field
    """

    """ name of field in DB """
    field: str

    """ textual label to identify this field """
    label: str

    """ use of the field NB under review probably wont support in grid editing """
    mode: FieldType = FieldType.STRING

    """ render width in chars """
    width: int

    def __init__(self, field: str, label: str, mode: FieldType):
        self.field = field
        self.mode = mode
        self.label = label


class Filter:
    """ used to bundle together values that are used to create a where clause
        Parameters:
            name(str): The name or label of the filter
            field(str): the name of the db field
            op(str): operation could be = or like for example
            clr_val(str): value to use when clearing filter input (a default)
            pre_val(str): a string to add before the filter value (could be ( or % )
            post_val(str): a string to add after the filter value (could be ) or % )

        optional members that can be set after creation:
            select(dict): key value pairs to use in a html select
    """
    clrVal: str  # value to clear form field to
    field: str  # name of the field to filter
    name: str  # name of the filter (used as a visual label)
    op: str  # operation like, = etc
    preVal: str  # string to add to filter value % or ( etc
    postVal: str  # after filter string
    select: {}  # a dict of id's to labels for a select tag

    def __init__(self, name, field, op, clr_val, pre_val='', post_val=''):
        self.name = name
        self.field = field
        self.op = op
        self.preVal = pre_val
        self.postVal = post_val
        self.select = {}
        self.value = ''
        self.clrVal = clr_val


class Grid:
    """ A "grid" like component, that represents a paginated record set of the
    results of an SQL query

        parameters:
            name(str): The name used to identify the grid
            csrf(str): token to help mitigate Cross Site Request Forgery

        Other properties set after creation:
            query(str): SQL fragment of the query before the where clause
            params(dict): key value pairs of parameters for the query
            rpp(int): Records Per Page to show
            page(int): The page to show
            fields(list[Field]): list of field object
            idField(str): which field uniquely identifies a record
            actions(list[str]): List of strings, add given action buttons to a grid
        """
    name: str
    query: str
    order: str
    params = {}
    rpp: int = 6
    page: int = 1
    record_count: int = 0
    fields: List[Field] = []
    idField: str = ''
    csrf: str = ''
    actions: List[str] = []

    def __init__(self, name: str, csrf: str):
        self.name = name
        self.csrf = csrf

    def populate(self, scope, parent: Element, id_prefix: str):
        """ populate the grid so the current page can be shown
        conn(Conn): connection object to the db
        parent(Element) the tag to add the grid to
        id-prefix(str): This is the prefix for the id's array that identifies all the pages lines this helps give
        a unique key value for every grid in the domain when saving their line ids in the session values table
        """
        conn = scope['state']['conn']
        # get the record count for the whole set (all pages)
        try:
            rs = conn.execute(text("with cq as ( " + self.query + " ) select count(*) count from cq"),
                              self.params).all()
        except Exception as e:
            print("with cq as ( " + self.query + " ) select count(*) count from cq", self.params)
            raise e

        self.record_count = rs[0][0]

        # having given filters chance to add a where clause the order by clause can be added
        if hasattr(self, 'order'):
            self.query += ' order by ' + self.order

        # calculate how many records in the current page is
        start = (self.page - 1) * self.rpp
        limit = " limit " + str(start) + "," + str(self.rpp)

        # fetch the current page of data
        try:
            rs = conn.execute(text(self.query + limit), self.params)
        except Exception as e:
            print(self.query + limit, file=sys.stderr)
            raise e

        rs = rs.mappings().all()
        frm = util.create_form(parent, self.name, self.csrf)
        # present the current page in an HTML table
        tab = frm.add_tag("table")
        tr = tab.add_tag("tr")
        for f in self.fields:
            if f.label:
                tr.add_tag('th', content=f.label)

        for _ in self.actions:
            tr.add_tag('th')

        line_vals = {}
        line = 0
        for r in rs:
            tr = tab.add_tag("tr")
            for f in self.fields:
                if f.label:
                    tr.add_tag("td", content=str(r[f.field]))
                if f.field == self.idField:
                    line_vals[id_prefix + '[' + str(line) + ']'] = r[f.field]
            for a in self.actions:
                atd = tr.add_tag('td')
                atd.add_tag('button',
                            {'name': 'action', "type": "submit", "value": a + str(line)}, content=a)
            line += 1

        lp = int(ceil(self.record_count / self.rpp)) + 1
        for i in range(1, lp):
            if abs(i - self.page) < 6 or i == 1 or i == lp - 1 or i % 10 == 0:
                attrib = {"name": "page", "class": "btn-class", "type": "submit", "value": str(i)}
                if self.page == i:
                    attrib["disabled"] = ""
                frm.add_tag('button', attrib, content=str(i))

        return line_vals


# util functions for grids
def get_filter_vals(globs, formname, filt, post):
    """ Gets the filter value (for a grid) from a POST request
    Parameters:
        globs(dict): persistent key value pairs for the current page
        formname(str): name of the filter form
        filt(List(Filter)): list of filter objects representing each item in the form
        post(dict): dict of POSTed values
    """
    for f in filt:
        globs[formname + f.field] = post.get('filter_' + f.field, [''])[0]


def clear_filter_vals(globs, formname, filt):
    """clears or resets the filters to their default values
    Parameters:
        globs(dict): persistent key value pairs for the current page
        formname(str): the name of the filter form
        filt(List(Filter)): list of filter object for the form
    """
    for f in filt:
        globs[formname + f.field] = f.clrVal


def create_filter_form(parent, form_name, grid, filt, globs, cls=None):
    """
    creates a form to allow filtering of a grid
    Parameters:
        parent(Element): The parent tag to attach the form to
        form_name(str): The name of the form
        grid(grid.Grid): The name of the grid that is being filtered
        filt(List(Filter)): list of filter object to construct the form from
        globs(dict): persistent key value pairs for the current page
        cls(str): option CSS class for the form

    Returns:
        The form tag that has been added to the parent
    """
    frm = util.create_form(parent, form_name, globs['csrf'])
    att = {}
    if cls:
        att = {'class': cls}
    tab = frm.add_tag('table', att)
    tr = tab.add_tag('tr')
    for f in filt:
        tr.add_tag('th', content=f.name)

    tr.add_tag('th')  # space for buttons
    tr.add_tag('th')
    tr = tab.add_tag('tr')

    for f in filt:
        td = tr.add_tag('td')
        if not f.select:
            td.add_tag('input', {'type': 'text', 'name': 'filter_' + f.field,
                                 'value': globs.get(form_name + f.field, '')})
        else:
            sel = td.add_tag('select', {'name': 'filter_' + f.field})
            for k, v in f.select.items():
                attr = {'value': k}
                if k == globs.get(form_name + f.field):
                    attr['selected'] = None
                sel.add_tag('option', attr, v)

    td = tr.add_tag('td')
    td.add_tag('input', {'type': 'submit', 'name': 'action', 'value': 'filter'})
    td = tr.add_tag('td')
    td.add_tag('input', {'type': 'submit', 'name': 'action', 'value': 'clear'})

    wh = False
    for f in filt:
        if globs.get(form_name + f.field, '') != '':
            if not wh:
                wh = True
                grid.query += ' where '
            grid.params[f.field] = f.preVal + globs[form_name + f.field] + f.postVal
            grid.query += f.field + ' ' + f.op + ' %(' + f.field + ')s and '
    if wh:
        grid.query = grid.query[:-4]

    return frm

