import datetime
import io
import os
import pprint
from functools import wraps
from http import cookies
from uuid import uuid4
from xml.dom.minidom import getDOMImplementation, Element
from xml.dom.minidom import parseString

from sqlalchemy import text

from app.lib import settings


def add_method(cls):
    """ decorator to add a method to an existing class """

    def decorator(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            return func(*args, **kwargs)

        setattr(cls, func.__name__, wrapper)
        return func  # returning func means func can still be used normally

    return decorator


@add_method(Element)
def add_tag(self, tag, attrib=None, content=None):
    """ convienience method added to the Element class """
    if attrib is None:
        attrib = []
    t = self.ownerDocument.createElement(tag)
    if content is not None:
        if tag.lower() == 'script':
            # use html comment to protect script from xml exporter
            t.appendChild(self.ownerDocument.createComment("//<![CDATA[\n" + content + "\n//]]>"))
        else:
            t.appendChild(self.ownerDocument.createTextNode(str(content)))

    for a in attrib:
        t.setAttribute(a, attrib[a])
        if a == 'name':
            t.setAttribute('id', attrib[a])
            t.setIdAttribute('id')
        if a == 'id':
            t.setAttribute('name', attrib[a])
            t.setIdAttribute('id')
    self.appendChild(t)
    return t


async def read_body(receive):
    """
    Read and return the entire (request) body from an incoming ASGI message.
    """
    body = b''
    more_body = True

    while more_body:
        message = await receive()
        body += message.get('body', b'')
        more_body = message.get('more_body', False)

    return body


def create_page(title, scope):
    """ convienience function to create the xhtml document with title and body """
    impl = getDOMImplementation()
    dt = impl.createDocumentType(
        "html", "-//W3C//DTD XHTML 1.0 Transitional//EN",
        "http://www.w3.org/TR/xhtml1/DTD/xhtml1-transitional.dtd")
    dom = impl.createDocument("http://www.w3.org/1999/xhtml", "html", dt)
    html = dom.documentElement
    html.setAttribute("xmlns", "http://www.w3.org/1999/xhtml")

    head = dom.createElement("head")
    head.add_tag("link", {"href": settings.APP_ROOT + "/style.css",
                          "rel": "stylesheet",
                          "type": "text/css"})
    html.appendChild(head)

    t = dom.createElement('title')
    t.appendChild(dom.createTextNode(title))
    head.appendChild(t)

    body = dom.createElement("body")
    html.appendChild(body)
    banner = import_template(body, 'header.xml')

    usr = scope['state'].get('user', None)
    outlink = dom.getElementById('outlink')
    if outlink:
        if usr:
            expires_tag = dom.getElementById('expires')
            if expires_tag:
                banner.add_tag('script', {}, """
  var exp = new Date('""" + str(usr['expires']) + """').getTime();
  var x = setInterval(function() {
      var now = new Date().getTime();
      var distance = exp - now;
      var minutes = Math.floor((distance % (1000 * 60 * 60)) / (1000 * 60));
      var seconds = Math.floor((distance % (1000 * 60)) / 1000);

      if (distance < 0) {
        clearInterval(x);
        window.location.href = '""" + settings.LOGOUT_PAGE + """'
      } else {
        document.getElementById("expires").innerHTML = "Page expires in " + minutes + ":" 
              + seconds.toString().padStart(2, "0");
      }
  }, 1000);
                """)
        else:
            outlink.removeChild(outlink.firstChild)
            outlink.add_tag('a', {'href': settings.APP_ROOT + os.sep + settings.LOGIN_PAGE}, 'Log In')

    return dom, body


async def render_page(scope, send, dom):
    raw = '\n'.join([line for line in dom.toprettyxml(indent='    ').split('\n') if line.strip()]).encode('utf-8')
    headers = [
        (b'content-type', b'text/html'),
        (b'content-length', str(len(raw)).encode())
    ]
    if "extra_headers" in scope['state']:
        for eh in scope['state']["extra_headers"]:
            headers.append(eh)

    await send({
        'type': 'http.response.start',
        'status': 200,
        'headers': headers
    })
    await send({
        'type': 'http.response.body',
        'body': raw
    })


def import_template(parent, templatename):
    """ imports a static template in raw xhtml into a DOM and insert it as a child
        of the specified node, the template MUST have exactly one root node
        once imported the template can be manipulated with minidom as part
        of the DOM it's been parented to.
        Parameters:
            parent(Node):       an xml node in a DOM to attach the template to
            templatename(str):  path and filename relative to the templates directory
        """
    template_path = os.path.join('app', 'templates', templatename)
    try:
        with open(template_path) as f:
            raw = f.read().strip()
            template = parseString(raw)
    except FileNotFoundError:
        raise FileNotFoundError("Template file not found: {}".format(template_path))

    nodes = template.getElementsByTagName("*")
    for node in nodes:
        node.ownerDocument = parent.ownerDocument
        if node.hasAttribute('id'):
            node.setIdAttribute('id')

    # Modify URLs in all elements that have a URL attribute
    for elem in template.getElementsByTagName('*'):
        for attr_name in elem.attributes.keys():
            # none local URL's specified with protocol (http / https)
            # local URL's should start with a /
            # but is actually relative to APP_ROOT
            old_url = elem.attributes[attr_name].value
            if not old_url.lower().startswith('http:') and \
                    not old_url.lower().startswith('https:'):

                if attr_name in ['href', 'src', 'action', 'data', 'srcset', 'usemap']:
                    new_url = settings.APP_ROOT + old_url
                    elem.attributes[attr_name].value = new_url

    de = template.documentElement
    parent.appendChild(template.documentElement)
    return de


def msg(parent, content, error=False):
    """ adds a message box that can be info (green) or an error message (red)
    Parameters
    parent(Element):    tag to add the message to
    text(str):          The text of the message
    error(bool):        Should it use the info or alert CSS
    """
    cls = "info"
    if error:
        cls = "alert"
    mesg = parent.add_tag("div", {"class": cls})
    mesg.add_tag("span", {"class": "closebtn", "onclick": "this.parentElement.style.display='none';"}, content='x')
    mesg.add_tag("span", content=content)


def render_plot(plot, parent):
    """ Save a matplotlib chart as a PNG image
    Parameters:
        plot(pyplot):       The plot to render as an image
        parent(Element):    The parent tag to add the IMG tag to.
    """
    buf = io.BytesIO()
    plot.savefig(buf, format='svg', bbox_inches='tight')
    buf.seek(0)

    tag = parseString(buf.getvalue().decode('UTF-8')).documentElement
    parent.appendChild(tag)
    return tag


def create_form(parent, name, csrf):
    """ creates a form embedding CSRF token and form name
    Parameters:
        parent: parent tag for the form
        name: the name attribute for the form
        csrf: the csrf token to use to help mitigate Cross Site Request Forgery
    """
    frm = parent.add_tag("form", {"name": name, "method": "POST"})
    frm.add_tag("input", {'type': 'hidden', 'name': 'csrf', 'value': csrf})
    frm.add_tag("input", {'type': 'hidden', 'name': 'form', 'value': name})
    return frm


def create_select(parent, name, items, chosen):
    """
    creates an HTML select tag and its options to select from
    parameters:
    parent(Element):            html node to attach the select to
        name(str):              the name attribute
        items(list of dict):    list of dicts where "id" is returned and value is displayed for each option
        chosen(str):            which is any option should be preselected
    :return: the selection node (can be useful for chaining)
    """
    sel = parent.add_tag('select', {'name': name})
    for item in items:
        item['id'] = str(item['id'])
        attrib = {'value': item['id']}
        if str(item['id']) == str(chosen):
            attrib['selected'] = None
        sel.add_tag('option', attrib, str(item['value']))
    return sel


def add_to_dict(adict, key, values):
    """ create or add to a key """
    if key not in adict:
        adict[key] = [values]
    else:
        adict[key].append(values)


def renew_session(scope):
    sessval = str(uuid4())
    scope['state']['session'] = sessval
    add_to_dict(scope['state'], 'extra_headers', (b"Set-Cookie", ("sess = " + sessval + "; path = /").encode('utf-8')))
    conn = scope['state']['conn']
    conn.execute(text("update " + settings.DB_CRED['USERTAB'] + " set session = :sess where id = :id"),
                 {'sess': sessval, 'id': scope['state']['user']['id']})
    scope['state']['user']['session'] = sessval
    conn.commit()


def session_cookie(scope):
    """ get the session cookie and add the session value to scope
        if a session cookie isn't found create a new one via _extra_header """
    headers_dict = {name.decode('utf-8').lower(): value.decode('utf-8')
                    for name, value in scope['headers']}
    cookie_value = headers_dict.get('cookie')
    sessval = str(uuid4())

    if cookie_value is not None:
        cookie = cookies.SimpleCookie()
        try:
            cookie.load(cookie_value)
            sessval = cookie.get('sess', None).value
        except cookies.CookieError:
            print("cookie error")
        if sessval is None or sessval == '':
            sessval = str(uuid4())
    else:
        add_to_dict(scope['state'], 'extra_headers',
                    (b"Set-Cookie", ("sess = " + sessval + "; path = /").encode('utf-8')))
    scope['state']['session'] = sessval


def get_page_globals(scope):
    """ fetch page globals from session these are persistent variable per page
    Parameters:
        scope: the request scope
    Returns(dict): A dictionary of the globals for the requested domain
    """
    conn = scope['state']['conn']
    uid = scope['state']['user']['id']
    domain = scope['path']
    rs = conn.execute(text("select name, value from " + settings.DB_CRED['SESSVARTAB'] +
                           " where uid=:uid and domain=:domain "),
                      {'uid': uid, 'domain': domain}).all()
    return dict(rs)


def save_page_globals(scope, globs):
    """ saves/updates page globals to the users session
    Parameters:
        scope: request scope
        globs(dict): the key/values to save/update
    """
    uid = scope['state']['user']['id']
    domain = scope['path']
    data = [{'uid': uid, 'domain': domain, 'name': key, 'value': str(value)} for key, value in globs.items()]

    conn = scope['state']['conn']
    conn.execute(text('delete from ' + settings.DB_CRED['SESSVARTAB'] +
                      ' where uid = :uid and domain = :domain'), {'domain': domain, 'uid': uid})
    conn.execute(text('insert into ' + settings.DB_CRED['SESSVARTAB'] +
                      ' (uid,domain,name,value) values(:uid, :domain, :name, :value)'), data)
    conn.commit()
    # conn.close()
    return


def clear_all_globals(scope):
    """ delete all users page globals (user must be logged in) """
    conn = scope['state']['conn']
    conn.execute(text('delete from ' + settings.DB_CRED['SESSVARTAB'] + ' where uid = :uid '),
                 {'uid': scope['state']['user']['id']})
    conn.commit()


def get_post_var(parsed, varname):
    return parsed.get(varname, [b''])[0].decode("utf-8")


def dump_scope(parent, scope):
    pp = ''
    for sc in scope:
        pp += str(sc) + ' = ' + pprint.pformat(scope[sc], indent=4) + '\n'
    parent.add_tag('pre', {'style': 'font-size: 0.6em;'}, pp)
