import datetime
import importlib
import os.path
from traceback import format_exc

from sqlalchemy import create_engine, text, NullPool

from app.lib import util, settings


async def app(scope, receive, send):
    if scope['type'] == 'lifespan':
        while True:
            message = await receive()
            if message['type'] == 'lifespan.startup':
                print('--- STARTUP ----')
                # DB engaine shared amongst scopes
                scope['state']['db'] = create_engine(settings.DB_CRED['CONSTR'], poolclass=NullPool)
                await send({'type': 'lifespan.startup.complete'})
            elif message['type'] == 'lifespan.shutdown':
                print('--- SHUTDOWN ----')
                scope['state']['db'].dispose()
                await send({'type': 'lifespan.shutdown.complete'})
                return

    assert scope['type'] == 'http'

    # remove settings.APP_ROOT from path
    # translates /app/path/route to /route

    if scope['path'].startswith(settings.APP_ROOT):
        scope['path'] = scope['path'][len(settings.APP_ROOT):]
    if scope['path'] == '':
        scope['path'] = '/'

    # turn URL into file path
    path = 'app/pages' + scope.get('path', '/')
    if scope.get('path', '/') == '/':
        path += 'index'

    path = path.replace('/', os.path.sep)   # *sigh*

    # no endpoint for path check for (and serve) a static file
    if not os.path.exists(path + ".py"):
        path = scope['path']
        typ = path[-4:]
        headers = []

        if typ == ".png":
            headers.append((b"Content-Type", b"image/png"))
        if typ == ".ico":
            headers.append((b"Content-Type", b"image/x-icon"))
        if typ == ".css":
            headers.append((b"Content-Type", b"text/css"))

        headers.append((b'Cache-Control', b'max-age=31536000'))

        try:
            with open('app' + path, 'rb') as fp:
                data = fp.read()

                headers.append((b'Content-Length', str(len(data)).encode('utf8')))
                await send({
                    'type': 'http.response.start',
                    'status': 200,
                    'headers': headers
                })
                await send({
                    'type': 'http.response.body',
                    'body': data
                })
                return
        # no endpoint or static file to serve - 404
        except FileNotFoundError:
            dom, body = util.create_page("404 Not Found", scope)
            body.add_tag('div', {}, 'Page Not Found ')
            body.add_tag('a', {'href': '/'}, "Main Page")
            await util.render_page(scope, send, dom)
            return

    util.session_cookie(scope)

    # db connection for this scope
    scope['state']['conn'] = scope['state']['db'].connect()

    # get user if logged in
    conn = scope['state']['conn']
    urs = conn.execute(text("SELECT * FROM " + settings.DB_CRED['USERTAB'] +
                            " WHERE session = :uid and session is not NULL "),
                       {'uid': scope['state']['session']})
    urs = urs.mappings().first()
    if urs:
        scope['state']['user'] = dict(urs)
    else:
        scope['state']['user'] = None

    # if page requires login redirect if not logged in or refresh user session token
    if scope['path'] not in settings.PUBLIC_PAGES:
        if urs is None:
            await send({
                'type': 'http.response.start',
                'status': 302,
                'headers': [(b'Location', (settings.APP_ROOT+'/'+settings.LOGIN_PAGE+
                                           '?redir='+scope['path']).encode('utf-8'))]
            })
            await send({
                'type': 'http.response.body',
                'body': b''
            })
            return
        else:
            # mitigate session fixation, probably needn't be done *every* request...
            util.renew_session(scope)

    # if logged in check if session expired
    if scope['state']['user']:
        expires = scope['state']['user']['expires'] - datetime.datetime.now()
        if expires.days < 0 or expires.seconds < 0:
            await send({
                'type': 'http.response.start',
                'status': 302,
                'headers': [(b'Location', (settings.APP_ROOT+'/'+settings.LOGOUT_PAGE).encode('utf-8'))]
            })
            await send({
                'type': 'http.response.body',
                'body': b''
            })
            return

    # activity resets idle timer
    if scope['state']['user']:  # update expiry time
        expires = datetime.datetime.now() + datetime.timedelta(seconds=settings.MAX_SESS_TIME)
        scope['state']['user']['expires'] = expires
        conn.execute(text("update " + settings.DB_CRED['USERTAB'] + " set expires = :expires where id = :id"),
                     {'expires': expires, 'id': scope['state']['user']['id']})

        conn.commit()

    # turn path into module path
    path = path.replace(os.path.sep, '.')

    # catching broad exceptions as there could be any reason for an exeception in an endpoint.
    # noinspection PyBroadException
    try:
        page = importlib.import_module(path)
    except Exception:
        # TODO stacktrace to stdout and less specific message to response
        # dependant on IS_DEV ?
        dom, body = util.create_page("Exception importing route", scope)
        body.add_tag('pre', {}, format_exc())
        await util.render_page(scope, send, dom)
        scope['state']['conn'].close()
        return

    # noinspection PyBroadException
    try:
        await page.index(scope, receive, send)
    except Exception:
        # TODO stacktrace to stdout and less specific message to response
        dom, body = util.create_page("Exception", scope)
        body.add_tag('pre', {}, format_exc())
        await util.render_page(scope, send, dom)

    # close this scopes DB connection
    scope['state']['conn'].close()
